// Datei XmlJStd.java
/* ------------------------------------------------------------------------
Dieser Modul enthaelt Methoden zum Bearbeiten von XML-Dateien und
benutzt dazu nur die Standardbibliothek von Java 6.

Die wichtigen Methoden:
lies             Liest eine XML-Datei und erzeugt daraus ein
                 entsprechendes DOM-Document-Objekt. Die Datei
                 wird NICHT auf Gueltigkeit geprueft.
                 Eingelesen wird mit  DOMImplementationLS.
lies2            Wie lies, aber
                 eingelesen wird mit DocumentBuilderFactory.
liesVD           Wie lies2, aber die Datei wird auf Gueltigkeit
                 relativ zu ihrer DTD geprueft.
liesVS           Wie lies2, aber die Datei wird auf Gueltigkeit
                 relativ zu einem XML-Schema geprueft.
istGueltig       Prueft, ob ein DOM-Document-Objekt gueltig ist
                 relativ zu seiner DTD.
schreib          Schreibt ein DOM-Document-Objekt im XML-Format
                 (und "schoen eingerueckt") in eine Datei.

Weniger wichtige Methoden:
gibTextDateiAus  Gibt eine beliebige Textdatei zur
                 Standardausgabe aus
gibLesbarAus     Gibt ein DOM-Document-Objekt in einer selbstgestrickten
                 Form ("als Einrueck-Baum") zur Standardausgabe aus
---------------------------------------------------------------------------
Ausserdem enthaelt dieser Modul eine main-Methode, die einige der
oben erwaehnten Methoden ein bisschen testet.
------------------------------------------------------------------------ */
import java .io               .File;
import java .io               .FileInputStream;
import java .io               .FileReader;
import java .io               .FileWriter;
import javax.xml.parsers      .DocumentBuilderFactory;
import javax.xml.parsers      .DocumentBuilder;
import javax.xml.validation   .SchemaFactory;
import javax.xml.validation   .Schema;
import javax.xml              .XMLConstants;
import javax.xml.validation   .Validator;
import javax.xml.transform.dom.DOMSource;

import org  .w3c.dom          .Document;
import org  .w3c.dom          .Node;
import org  .w3c.dom          .NodeList;
import org  .w3c.dom          .Element;
import org  .w3c.dom          .Comment;
import org  .w3c.dom          .Text;
import org  .w3c.dom          .ProcessingInstruction;
import org  .w3c.dom          .DOMConfiguration;
import org  .w3c.dom          .DOMStringList;
import org  .w3c.dom          .DOMErrorHandler;
import org  .w3c.dom          .DOMError;

import org  .w3c.dom.bootstrap.DOMImplementationRegistry;
import org  .w3c.dom.ls       .DOMImplementationLS;
import org  .w3c.dom.ls       .LSParser;
import org  .w3c.dom.ls       .LSSerializer;
import org  .w3c.dom.ls       .LSInput;

import org  .xml.sax          .ErrorHandler;
import org  .xml.sax          .SAXParseException;
import org  .xml.sax.helpers  .DefaultHandler;
//import org  .xml.sax.ext      .DefaultHandler2;

// Abkuerzungen fuer die Konstanten ATTRIBUTE_NODE bis TEXT_NODE:
import static org.w3c.dom.Node.*;

public class XmlJStd {
   // ---------------------------------------------------------------------
   static public Document lies(final String pfadXml) {
      // Liest die XML-Datei pfadXml ein, macht daraus ein DOM-Document-
      // Objekt und liefert dies als Ergebnis.
      // Falls eine Ausnahme auftritt, wird eine Fehlermeldung ausgegeben
      // und null als Ergebnis geliefert.
      // Beim Einlesen wird die XML-Datei NICHT auf Gueltigkeit geprueft
      // (siehe auch die Methoden lies2, liesVD und liesVS).

      try {
         DOMImplementationRegistry registry   =
            DOMImplementationRegistry.newInstance();
         DOMImplementationLS       impl       =
            (DOMImplementationLS) registry.getDOMImplementation("LS");

         short    mode       = DOMImplementationLS.MODE_SYNCHRONOUS;
         String   schemaType = null;
         LSParser parser     = impl.createLSParser(mode, schemaType);

         // Laut Dokumentation muessten CDATASections und Entity Referecnces
         // standardmaessig erhalten bleiben. Tatsaechlich muss man das
         // (im September 2007) ausdruecklich konfigurieren, etwa so:
         DOMConfiguration
                  config     = parser.getDomConfig();
         config.setParameter("cdata-sections", true);
         config.setParameter("entities",       true);

         Document domDoc     = parser.parseURI(pfadXml);
         return   domDoc;
      } catch (Exception ex) {
         meldeAusnahme(ex, "lies");
         return null; // Anstelle eines DOM-Document-Objekts
      }
   } // lies
   // ---------------------------------------------------------------------
   static public Document lies2(final String pfadXml) {
      // Liest die XML-Datei pfadXml ein, macht daraus ein DOM-Document-
      // Objekt und liefert dies als Ergebnis.
      // Falls eine Ausnahme auftritt, wird eine Fehlermeldung ausgegeben
      // und null als Ergebnis geliefert.
      // Beim Einlesen wird die XML-Datei NICHT auf Gueltigkeit geprueft
      // (siehe auch die Methoden lies, liesVD und liesVS).

      try {
         FileInputStream fis = new FileInputStream(pfadXml);

         // Ein DocumentBuilderFactory-Objekt erzeugen (mit der
         // Klassenmethode (static method) newInstance:
         DocumentBuilderFactory
         documentBuilderFactory = DocumentBuilderFactory.newInstance();
         // Das Fabrikobjekt konfigurieren:
         documentBuilderFactory.setIgnoringComments(false);
         documentBuilderFactory.setNamespaceAware  (true);
         documentBuilderFactory.setValidating      (false);
         // Die Datei pfadXml wird beim Einlesen NICHT validiert, nur
         // auf Wohlgeformtheit geprueft.

         // Ein DocumentBuilder-Objekt erzeugen:
         DocumentBuilder
         documentBuilder = documentBuilderFactory.newDocumentBuilder();

         // Die Objektmethode parse liest eine Datei ein und versucht,
         // aus dem Inhalt ein DOM-Document-Objekt zu erzeugen:
         Document domDoc = documentBuilder.parse(fis);

         // Das Document-Objekt wird (als "Einrueck-Baum") ausgegeben:
         // gibLesbarAus(doc);

         return domDoc;
      } catch (Exception ex) {
         meldeAusnahme(ex, "lies2");
         return null; // Anstelle eines DOM-Document-Objekts
      }
   } // lies2
   // ---------------------------------------------------------------------
   static public Document liesVD(final String pfadXml) {
      // Liest die XML-Datei pfadXml ein, macht daraus ein DOM-Document-
      // Objekt und liefert dies als Ergebnis.
      // Falls eine Ausnahme auftritt, wird eine Fehlermeldung ausgegeben
      // und null als Ergebnis geliefert.
      // Beim Einlesen wird die XML-Datei VALIDIERT gegen ihre DTD
      // (deshalb "VD"). Siehe auch die Methoden lies, lies2 und liesVS.

      // Ein geeignetes ErrorHandler-Objekt (einer anonymen Erweiterung
      // der Klasse DefaultHandler) erzeugen:
      ErrorHandler meinEH = new DefaultHandler() {
         public void fatalError(SAXParseException e) {
            meldeFehlerD(e, "fatalError", pfadXml);
         }
         public void error(SAXParseException e) {
            meldeFehlerD(e, "error", pfadXml);
         }
         public void warning(SAXParseException e) {
            meldeFehlerD(e, "warning", pfadXml);
         }
      };

      try {
         FileInputStream fis = new FileInputStream(pfadXml);

         // Ein DocumentBuilderFactory-Objekt erzeugen:
         DocumentBuilderFactory
         documentBuilderFactory = DocumentBuilderFactory.newInstance();
         // Das Fabrikobjekt konfigurieren:
         documentBuilderFactory.setIgnoringComments(false);
         documentBuilderFactory.setNamespaceAware  (true);
         documentBuilderFactory.setValidating      (true);
         // Validiert gegen die DTD des Dokuments!

         // Ein DocumentBuilder-Objekt erzeugen und mit einem
         // ErrorHandler-Objekt ausruesten:
         DocumentBuilder
         documentBuilder = documentBuilderFactory.newDocumentBuilder();
         documentBuilder.setErrorHandler(meinEH);

         // Die Objektmethode parse liest eine Datei ein und versucht,
         // aus dem Inhalt ein DOM-Document-Objekt zu erzeugen:
         Document domDoc = documentBuilder.parse(fis);

         // Das Document-Objekt wird (als "Einrueck-Baum") ausgegeben:
         // gibLesbarAus(doc);

         return domDoc;
      } catch (Exception ex) {
         meldeAusnahme(ex, "liesVD");
         return null; // Anstelle eines Document-Objekts
      }
   } // liesVD
   // ---------------------------------------------------------------------
   static public Document liesVS(final String pfadXml,
                                 final String pfadXsd)
   {
      // Liest die XML-Datei pfadXml ein, macht daraus ein DOM-Document-
      // Objekt und liefert dies als Ergebnis.
      // Falls eine Ausnahme auftritt, wird eine Fehlermeldung ausgegeben
      // und null als Ergebnis geliefert.
      // Beim Einlesen wird die XML-Datei VALIDIERT gegen das XML-Schema
      // in der Datei pfadXsd (deshalb "VS"). Siehe auch die Methoden lies
      // lies2 und liesVD.

      // Ein geeignetes ErrorHandler-Objekt (einer anonymen Erweiterung
      // der Klasse DefaultHandler) erzeugen:
      ErrorHandler meinEH = new DefaultHandler() {
         public void fatalError(SAXParseException e) {
            meldeFehlerS(e, "fatalError", pfadXml);
         }
         public void error(SAXParseException e) {
            meldeFehlerS(e, "error", pfadXml);
         }
         public void warning(SAXParseException e) {
            meldeFehlerS(e, "warning", pfadXml);
         }
      };

      try {
         FileInputStream fis = new FileInputStream(pfadXml);

         // Ein DocumentBuilderFactory-Objekt erzeugen:
         DocumentBuilderFactory
         documentBuilderFactory = DocumentBuilderFactory.newInstance();

         SchemaFactory
         schemaFactory          =
            SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
         File          file     = new File(pfadXsd);
         Schema        schema   = schemaFactory.newSchema(file);

         // Das Fabrikobjekt konfigurieren:
         documentBuilderFactory.setIgnoringComments(false);
         documentBuilderFactory.setNamespaceAware  (true);
         documentBuilderFactory.setValidating      (false);
         documentBuilderFactory.setSchema          (schema);
         // Damit das Schreiben eines validierenden Parsers nicht zu
         // einfach ist :-) muss man hier setValidating(false) angeben!
         // setValidating(true) wuerde bewirken, dass gegen eine DTD
         // validiert wird. Hier soll aber gegen eine XML-Schema-Datei
         // validiert werden.

         // Ein DocumentBuilder-Objekt erzeugen und mit einem
         // ErrorHandler-Objekt ausruesten:
         DocumentBuilder
         documentBuilder = documentBuilderFactory.newDocumentBuilder();
         documentBuilder.setErrorHandler(meinEH);

         // Die Objektmethode parse liest eine Datei ein und versucht,
         // aus dem Inhalt ein DOM-Document-Objekt zu erzeugen:
         Document domDoc = documentBuilder.parse(fis);

         // Das Document-Objekt wird (als "Einrueck-Baum") ausgegeben:
         // gibLesbarAus(doc);

         return domDoc;
      } catch (Exception ex) {
         meldeAusnahme(ex, "liesVS");
         return null; // Anstelle eines Document-Objekts
      }
   } // liesVS
   // ---------------------------------------------------------------------
   static void meldeFehlerD(SAXParseException e, String art, String pfadXml) {
      // art ist gleich "warning", "error" oder "fatalError". Dazu wird
      // ein passender unbestimmter Artikel (eine oder ein) ermittelt:
      String artikel = art.equals("warning") ? "eine" : "ein";
      printf("-------------------------------------------------%n");
      printf("Beim Validieren gegen eine DTD wurde%n");
      printf("%s *%s* gemeldet:%n", artikel, art);
      printf("Datei:   %s, Zeile %d, Spalte %d%n",
         pfadXml, e.getLineNumber(), e.getColumnNumber());
      printf("Meldung: %s%n%n", e.getMessage());
   } // meldeFehlerD
   // ---------------------------------------------------------------------
   static void meldeFehlerS(SAXParseException e, String art, String pfadXml) {
      // art ist gleich "warning", "error" oder "fatalError". Dazu wird
      // ein passender unbestimmter Artikel (eine oder ein) ermittelt:
      String artikel = art.equals("warning") ? "eine" : "ein";
      printf("-------------------------------------------------%n");
      printf("Beim Validieren gegen ein XML-Schema wurde%n");
      printf("%s *%s* gemeldet:%n", artikel, art);
      printf("Datei:   %s, Zeile %d, Spalte %d%n",
         pfadXml, e.getLineNumber(), e.getColumnNumber());
      printf("Meldung: %s%n%n", e.getMessage());
   } // meldeFehlerS
   // ---------------------------------------------------------------------
   static boolean istGueltigVD(Document domDoc) {
      // Validiert domDoc gegen seine Dokumenten-Typ-Definition (DTD).
      // Liefert true genau dann wenn domDoc gueltig ist.
      // Falls eine Ausnahme auftritt, wird eine Fehlermeldung ausgegeben
      // und false als Ergebnis geliefert. "VD" wie "validiert gegen DTD".
      //
      // Dazu wird domDoc in einen String "rausgeschrieben" und von dort
      // mit einem validierenden Parser "wieder reingelesen". Der Parser
      // muss mit einem DOMErrorHandler-Objekt (meinDomEH) konfiguriert
      // werden, damit in dieser Methode die Anzahl der (vom Parser er-
      // kannten) Fehler ermittelt werden kann (0 oder mehr als 0).

      // Methode meinDomEH.handleError soll die Anzahl der erkannten Fehler
      // festhalten. Sie kann aber nur auf final-Variablen in ihrer Umgebung
      // zugreifen. Darum wird hier nicht eine int-Variable namens anzFehler
      // vereinbart, sondern eine Reihungsvariable mit einer einzigen Kompo-
      // nente. In dieser Komponente anzFehler[0] werden die Fehler gezaehlt.
      final int[] anzFehler = {0};
      // ------------------------------------------------------------------
      // Ein geeignetes DOMErrorHandler-Objekt (einer anonymen Implemen-
      // tierung der Schnittstelle DOMErrorHandler) erzeugen.
      // Die Methode handleError kann nur auf final-Variablen der
      // umgebenden Methode istGueltigVD zugreifen (und diese Variablen
      // muessen textuell *vor* dem Objekt meinDomEH vereinbart werden).
      DOMErrorHandler meinDomEH = new DOMErrorHandler() {
         // Dieses Behandler-Objekt enthaelt nur eine Methode:
         public boolean handleError(DOMError dErr) {
            String  fehlerArt    = "Unbekannte Fehler-Art";
            boolean weiterMachen = false;
            int     severity     = dErr.getSeverity();

            switch (severity) {
               case DOMError.SEVERITY_WARNING:
                  fehlerArt    = "warning";
                  weiterMachen = true;
                  break;
               case DOMError.SEVERITY_ERROR:
                  fehlerArt    = "error";
                  weiterMachen = true;
                  break;
               case DOMError.SEVERITY_FATAL_ERROR:
                  fehlerArt    = "fatal error";
                  weiterMachen = false;
                  break;
            }

            printf("DOMError (%s): %s%n", fehlerArt, dErr.getMessage());
            anzFehler[0]++;

            // Nach Fehlern der Art WARNING oder ERROR soll der Parser
            // weitermachen, nach solchen der Art FATAL_ERROR nicht:
            return weiterMachen;
         }
      }; // meinDomEH
      // ------------------------------------------------------------------
      try {
         DOMImplementationRegistry registry =
            DOMImplementationRegistry.newInstance();
         DOMImplementationLS       impl     =
            (DOMImplementationLS) registry.getDOMImplementation("LS");

         // Das domDoc wird in den String strDoc geschrieben:
         LSSerializer     slizer  = impl.createLSSerializer();
         String           strDoc  = slizer.writeToString(domDoc);

         // Erzeuge einen synchronen, gegen eine DTD-validierenden Parser:
         short            mode    = DOMImplementationLS.MODE_SYNCHRONOUS;
         String           valArt  = "http://www.w3.org/TR/REC-xml"; // DTD
         LSParser         parser  = impl.createLSParser(mode, valArt);
         LSInput          input   = impl.createLSInput();

         // Den parser konfigurieren:
         DOMConfiguration config  = parser.getDomConfig();
         config.setParameter("validate",        true);
//       config.setParameter("schema-type",     "http://www.w3.org/TR/REC-xml");
//       config.setParameter("schema-location", "nameA.dtd");
         config.setParameter("error-handler",   meinDomEH);

         // Der String strDoc wird mit dem parser "eingelesen":
         input.setStringData(strDoc);
         Document         domDoc2 = parser.parse(input);
         // Aus der JavaDoc der Methode parse in der Schnittstelle LSParser:
         //
         // LSException - PARSE_ERR: Raised if the LSParser was unable to
         // load the XML document.
         //
         // Was bedeutet das? Sollte bei einem "normalen Parse-Error" eine
         // LSException geworfen werden oder nicht? Es wird KEINE geworfen!

         // Wenn das domDoc 0 Fehler enthaelt, ist es gueltig:
         return anzFehler[0] == 0;

      } catch (Exception ex) {
         meldeAusnahme(ex, "istGueltigVD");
         return false;   // Das domDoc ist nicht gueltig
      }
   } // istGueltigVD
   // ---------------------------------------------------------------------
   static public boolean istGueltigVS(Document domDoc, String pfadXsd) {
      // Validiert domDoc gegen das XML-Schem in der Datei pfadXsd.
      // Liefert true genau dann, wenn domDoc gueltig (engl. valid) ist.
      // Falls eine Ausnahme auftritt, wird eine Fehlermeldung ausgegeben
      // und false als Ergebnis geliefert.

      try {
         // Ein SchemaFactory-Objekt erzeugen:
         SchemaFactory schemaFactory =
         SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);

         // Ein Schema-Objekt und ein Validator-Objekt erzeugen:
         File      file      = new File(pfadXsd);
         Schema    schema    = schemaFactory.newSchema(file);
         Validator validator = schema.newValidator();

         // Validiert werden kann nicht das gesamte domDoc, sondern nur
         // sein documentElement (oder ein anderer Unterbaum):
         Element   docElem   = domDoc.getDocumentElement();
         // Die Methode validate erwartet ein DOMSource-Objekt:
         DOMSource domSource = new DOMSource(docElem);
         // Das docElem in der domSource wird auf Gueltigkeit geprueft.
         // Wird dabei ein Fehler festgestellt, so wird eine
         // Ausnahme (des Typs org.xml.sax.SAXParseException) geworfen:
         validator.validate(domSource);
         return true;    // Das domDoc ist gueltig

      } catch (Exception ex) {
         meldeAusnahme(ex, "istGueltigVS");
         return false;   // Das domDoc ist nicht gueltig
      }
   } // istGueltigVS
   // ---------------------------------------------------------------------
   static public void schreib(Document domDoc, String pfadXml) {
      // Das domDoc wird als XML-Datei mit dem Pfad pfadXml ausgegeben.

      final boolean TST = false;

      try {
         DOMImplementationRegistry registry =
            DOMImplementationRegistry.newInstance();
         DOMImplementationLS       impl     =
            (DOMImplementationLS) registry.getDOMImplementation("LS");

         LSSerializer     writer = impl.createLSSerializer();
         DOMConfiguration config = writer.getDomConfig();

         if (TST) {
            // Welche Parameter kann man im Objekt config konfigurieren?
            printf("************************************************%n");
            DOMStringList    stList = config.getParameterNames();
            printf("Erlaubte Parameter des DOMConfiguration-Objekts:%n");
            for (int i=stList.getLength()-1; i>=0; i--) {
               printf("parameter["+ i +"]: " + stList.item(i));
            }
            printf("************************************************%n");
         }

         if (config.canSetParameter("format-pretty-print", true)) {
            config.setParameter("format-pretty-print", true);
         }

         writer.writeToURI(domDoc, pfadXml);
      } catch (Exception ex) {
         meldeAusnahme(ex, "schreib");
      }
   } // schreib
   // ---------------------------------------------------------------------
   static public void gibTextDateiAus(String pfadTxt) {
      // Gibt die Textdatei mit dem Pfad pfadTxt zur Standardausgabe aus.
      try {
         FileReader fr = new FileReader(pfadTxt);
         int zeichen;
         while ((zeichen = fr.read()) != -1) printf("%c", zeichen);
         fr.close();
      } catch (Exception ex) {
         meldeAusnahme(ex, "gibTextDateiAus");
      }
   } // gibTextDateiAus
   // ---------------------------------------------------------------------
   static public void meldeAusnahme(Exception ex, String inMethode) {
      String ausnahme = ex.getClass().getName();
      printf("In XmlJStd.%s: %s%n", inMethode, ausnahme);
      printf("Message: %s%n", ex.getMessage());
   } // meldeAusnahme
   // ---------------------------------------------------------------------
   static public String nodeType(Node n) {
      // Liefert den Typ von n als String.
      // Die Konstanten ATTRIBUTE_NODE bis TEXT_NODE sind in der
      // Schnittstelle org.w3c.dom.Node vereinbart.
      switch (n.getNodeType()) {
         case              ATTRIBUTE_NODE: return "Attr: ";
         case          CDATA_SECTION_NODE: return "CDATASection: ";
         case                COMMENT_NODE: return "Comment: ";
         case               DOCUMENT_NODE: return "Document";
         case      DOCUMENT_FRAGMENT_NODE: return "DocumentFragment: ";
         case          DOCUMENT_TYPE_NODE: return "DocumentType: ";
         case                ELEMENT_NODE: return "Element: ";
         case                 ENTITY_NODE: return "Entity: ";
         case       ENTITY_REFERENCE_NODE: return "EntityReference: ";
         case               NOTATION_NODE: return "Notation: ";
         case PROCESSING_INSTRUCTION_NODE: return "ProcessingInstruction: ";
         case                   TEXT_NODE: return "Text: ";
         default:                          return "Unbekannter Typ";
      }
   } // nodeType
   // ---------------------------------------------------------------------
   static public void gibLesbarAus(Document domDoc) {
      // Gibt das domDoc in (hoffentlich) lesbarer Form
      // als "Einrueck-Baum" zur Standardausgabe aus.
      printf("XML-Version: %s%n", domDoc.getXmlVersion());
      printf("Encodierung: %s%n", domDoc.getXmlEncoding());
      printf("Standalone : %s%n", domDoc.getXmlStandalone());
      printf("%n");
      Element documentElement = domDoc.getDocumentElement();
      gibLesbarAusR(documentElement, "");
   } // gibLesbarAus
   // ---------------------------------------------------------------------
   static public void gibLesbarAusR(Node k, String einrueck) {
      // Gibt den Baum mit dem Wurzelknoten k aus (und ruft sich dazu
      // rekursiv auf). Die Baumstruktur wird durch Einrueckung dargestellt.
      // Uninteressante Knoten werden nicht ausgegebenen. Hinweis:
      // ATTRIBUTE (von Element-Knoten) erscheinen NICHT in diesem Baum.

      if (istUninteressant(k)) return;

       // Den Typ von k richtig eingerueckt ausgeben
      printf("%s%s", einrueck, nodeType(k));

      // Im Textinhalt mehrere transparente Zeichen durch ein ' ' ersetzen:
      String txt = normalisiere(k.getTextContent());

      // Bei Knoten bestimmter Typen werden zusaetzliche Infos ausgegeben:
      if        (k instanceof Element) {
         printf("<%s>",   ((Element) k).getTagName());
      } else if (k instanceof Comment) {
         printf("\"%s\"", ((Comment) k).getData());
      } else if (k instanceof Text   ) {
         printf("\"%s\"", txt);
      } else if (k instanceof ProcessingInstruction) {
         printf("%n%s   Fuer Programm: %s", einrueck, k.getNodeName());
         printf("%n%s   Instruktion:   %s", einrueck, k.getNodeValue());
      }
      printf("%n");

      // Alle Nachfolger-Knoten (Node-Objekte) von k ausgeben:
      NodeList kList = k.getChildNodes();
      for (int i=0; i<kList.getLength(); i++) {
         Node ni = kList.item(i);
         gibLesbarAusR(ni, einrueck+"   ");
      }
   } // gibLesbarAusR
   // ---------------------------------------------------------------------
   static boolean istUninteressant(Node n) {
      // Lifert true, wenn n ein Text-Knoten ist, der keine Nachfolger hat,
      // nur transparente Zeichen (engl. whitespace) und mind. ein
      // transparentes Zeichen enthaelt.

      if (n.getFirstChild()!=null) return false;

      String txt = n.getTextContent();
      if (txt.length()==0) return false;

      for (int i=0; i<txt.length(); i++) {
         if (!Character.isWhitespace(txt.charAt(i))) return false;
      }
      return true;
   } // istUninteressant
   // ---------------------------------------------------------------------
   static String normalisiere(String s) {
      // Liefert eine Kopie von s, in der jede Folge von transparenten
      // Zeichen (engl. sequence of white space characters) durch ein
      // einziges Blank ersetzt wurde.
      StringBuilder erg = new StringBuilder();
      boolean sindInnerhalb = false;
      for (int i=0; i<s.length(); i++) {
         char c = s.charAt(i);
         if (Character.isWhitespace(c)) {
            if (sindInnerhalb) continue;
            sindInnerhalb = true;
            erg.append(' ');
            continue;
         }
         erg.append(c);
         sindInnerhalb = false;
      }
      return erg.toString();
   } // normalisiere
   // ---------------------------------------------------------------------
   static public void main(String[] _) {
      printf("XmlJStd: Jetzt geht es los!%n");
      printf("********************************************************%n");

      String origXml = "daten01B.xml";
      String kopiXml = "dJStd01B.xml";
      String err_Xml = "daten01E.xml";
      String pfadXsd = "nameA.xsd";

      printf("Die Datei %s hat folgenden Inhalt:%n", origXml);
      printf("--------------------------------------------------------%n");
      gibTextDateiAus(origXml);
      printf("--------------------------------------------------------%n");
      printf("Die Datei %s wird in ein DOM-Objekt eingelesen!%n", origXml);
      Document domDoc = lies(origXml);
      printf("Das DOM-Objekt wird in die Datei %s ausgegeben!%n", kopiXml);
      schreib(domDoc, kopiXml);
      printf("Die Datei %s hat folgenden Inhalt:%n", kopiXml);
      printf("--------------------------------------------------------%n");
      gibTextDateiAus(kopiXml);
      printf("********************************************************%n");
      printf("Die Datei %s wird als DOM-Objekt eingelesen%n", err_Xml);
      printf("und das Objekt wird gegen das XML-Schema %s validiert.%n",
         pfadXsd);
      Document errDoc1 = lies(err_Xml);
      boolean  erg1    = istGueltigVS(errDoc1, pfadXsd);
      printf("Ergebnis von istGueltigVS: %b%n", erg1);
      printf("********************************************************%n");
      printf("Die Datei %s wird als DOM-Objekt eingelesen%n", err_Xml);
      printf("und das Objekt wird gegen seine DTD validiert.%n");
      Document errDoc2 = lies(err_Xml);
      boolean  erg2    = istGueltigVD(errDoc2);
      printf("Ergebnis von istGueltigVD: %b%n", erg2);
      printf("********************************************************%n");
      Document ergDoc3 = liesVD(err_Xml);
      printf("********************************************************%n");
      Document errDoc4 = liesVS(err_Xml, pfadXsd);
      printf("********************************************************%n");
      printf("XmlJStd: Das war's erstmal!%n");
   } // main
   // ---------------------------------------------------------------------
   // Mehrere Methoden mit kurzen Namen:
   static void printf(String f, Object... v) {System.out.printf(f, v);}
   // ---------------------------------------------------------------------
} // class XmlJStd
/* ------------------------------------------------------------------------
Die Ausgabe des Programms XmlJStd.java (2007-09-14):

XmlJStd: Jetzt geht es los!
********************************************************
Die Datei daten02B.xml hat folgenden Inhalt:
--------------------------------------------------------
<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<!-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Datei daten02B.xml:
   Wohlgeformt und gueltig (valid) nach personA.dtd und
   personA.xsd.
   Enthaelt einen Hinweis auf personA.dtd
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  -->
<!DOCTYPE person SYSTEM "personA.dtd">

<person>

  <name>
    <vor_name>Alan</vor_name>
    <nach_name>Turing</nach_name>
  </name>
  Normaler
  Text Nr 1
  <!--Kommentar zwischen 2 Texten-->
  Normaler Text Nr 2
  <beruf>Informatiker</beruf>
  <![CDATA[<beruf>Dozent</beruf>]]>
  <beruf>Mathematiker</beruf>
  Entity Ref.: &amp;lt; steht fuer &lt;
  <beruf land="England" zeit="WK2">Kryptograf</beruf>
  <?SuperEditor Zeilenwechsel ignorieren!?>
  &hi;

</person>
--------------------------------------------------------
Die Datei daten02B.xml wird in ein DOM-Objekt eingelesen!
Das DOM-Objekt wird in die Datei dJStd02B.xml ausgegeben!
Die Datei dJStd02B.xml hat folgenden Inhalt:
--------------------------------------------------------
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE person SYSTEM "personA.dtd">
<!-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Datei daten02B.xml:
   Wohlgeformt und gueltig (valid) nach personA.dtd und
   personA.xsd.
   Enthaelt einen Hinweis auf personA.dtd
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  -->
<person>
    <name>
        <vor_name>Alan</vor_name>
        <nach_name>Turing</nach_name>
    </name>
  Normaler
  Text Nr 1

    <!--Kommentar zwischen 2 Texten-->
  Normaler Text Nr 2
  <beruf>Informatiker</beruf><![CDATA[<beruf>Dozent</beruf>]]><beruf>Mathematiker</beruf>
  Entity Ref.: &amp;lt; steht fuer &lt;<beruf
        land="England" zeit="WK2">Kryptograf</beruf><?SuperEditor Zeilenwechsel ignorieren!?>&hi;</person>
********************************************************
XmlJStd: Das war's erstmal!
------------------------------------------------------------------------ */
